import curses
import enum
import threading
import time
import keyboard
import textwrap

def _init_curses():
    stdscr = curses.initscr()
    curses.noecho()
    curses.cbreak()
    stdscr.keypad(True)
    if curses.has_colors():
        curses.start_color()
        curses.init_pair(1, curses.COLOR_RED, curses.COLOR_BLACK) 
        curses.init_color(3, 13, 38, 79) # Dark Blue
        curses.init_pair(3, 3, curses.COLOR_BLACK)
        try:
            if curses.COLORS >= 256:
                curses.init_pair(2, 208, curses.COLOR_BLACK)
            else:
                curses.init_pair(2, curses.COLOR_RED, curses.COLOR_BLACK)
        except curses.error:
            curses.init_pair(2, curses.COLOR_RED, curses.COLOR_BLACK)
    return stdscr

class ExecutionStatus(enum.Enum):
    InProgress = 0
    CompletedSuccessfully = 1
    Failed = -1
    NoStatus = 2
    Warning = 3
    Error = 4
    Stdout = -2
    Stderr = -3

__window = _init_curses()

__steps = []

__windowpad = curses.newpad(65535, __window.getmaxyx()[1])
__pad = 0
__padlock = threading.RLock()

def __threaded_main_loop():
    global __pad, __offset_lock
    pause = False
    while True:
        if keyboard.is_pressed('up'):
            if pause: continue
            with __padlock:
                if __pad > 0:
                    __pad -= 1
            pause = True
        elif keyboard.is_pressed('down'):
            if pause: continue
            with __padlock:
                if __pad < 65535:
                    __pad += 1
            pause = True
        else:
            pause = False

def add_step(message: str, start_status = ExecutionStatus.InProgress):
    __steps.append(
        {
            'message': message,
            'status': start_status.value
        }
    )
    _output()
    return len(__steps) - 1

def finish_step(id: int, status: ExecutionStatus):
    __steps[id]['status'] = status.value
    _output()

_TO_REFRESH_OUTPUT = False

def _output():
    global __steps, __windowpad, __pad, __padlock

    __windowpad.clear()
    line = 0
    height, width = __window.getmaxyx()

    rows = height \
           -23   # Bottom four lines
    
    start_index = max(0, len(__steps) - rows)
    visible_steps = __steps[start_index:]
    
    for i, step in enumerate(visible_steps):
        if step['status'] == ExecutionStatus.InProgress.value:
            __windowpad.addstr(line, 0, f'[/] {step["message"]}')
        elif step['status'] == ExecutionStatus.CompletedSuccessfully.value:
            __windowpad.addstr(line, 0, f'[+] {step["message"]}')
        elif step['status'] == ExecutionStatus.Failed.value:
            __windowpad.addstr(line, 0, f'[-] {step["message"]}')
        elif step['status'] == ExecutionStatus.NoStatus.value:
            __windowpad.addstr(line, 0, f'[*] {step["message"]}')
        elif step['status'] == ExecutionStatus.Warning.value:
            __windowpad.addstr(line, 0, f'[!] {step["message"]}', curses.color_pair(2))
        elif step['status'] == ExecutionStatus.Error.value:
            __windowpad.addstr(line, 0, f'[!!] {step["message"]}', curses.color_pair(1) | curses.A_BOLD)
        elif step['status'] == ExecutionStatus.Stdout.value:
            lines = textwrap.wrap(step["message"].strip(), width - 1)
            for wrapped_line in lines:
                __windowpad.addstr(line, 0, f'    # {wrapped_line}', curses.color_pair(3) | curses.A_DIM)
                if len(visible_steps) - 1 == i:
                    continue
                if visible_steps[i+1]['status'] not in [ExecutionStatus.Stdout.value, ExecutionStatus.Stderr.value]:
                    line += 1
        elif step['status'] == ExecutionStatus.Stderr.value:
            lines = textwrap.wrap(step["message"].strip(), width - 1)
            for wrapped_line in lines:
                __windowpad.addstr(line, 0, f'    # {wrapped_line}', curses.color_pair(1) | curses.A_DIM)
                if len(visible_steps) - 1 == i:
                    continue
                if visible_steps[i+1]['status'] not in [ExecutionStatus.Stdout.value, ExecutionStatus.Stderr.value]:
                    line += 1
        
        line += 1

    __windowpad.addstr(line + 1, 0, '-' * (width - 1), curses.A_DIM)
    __windowpad.addstr(line + 2, 0, 'Press Ctrl+C to exit.', curses.A_DIM)
    if len(__steps) == 0:
        tmp = "0/0"
    else:
        end_index = min(len(__steps), start_index + rows)
        tmp = f"{start_index + 1}-{end_index}/{len(__steps)}"
    __windowpad.addstr(line + 3, 0, f'{tmp: ^{width}}', curses.A_DIM)

    with __padlock:
        __windowpad.refresh(__pad, 0, 0, 0, height - 1, width - 1)
    time.sleep(0.01)

def getscr():
    global __window, __windowpad
    return __windowpad


import atexit

def _cleanup_curses():
    global __window
    if __window:
        curses.nocbreak()
        __window.keypad(False)
        curses.echo()
        curses.endwin()

atexit.register(_cleanup_curses)

t = threading.Thread(target=__threaded_main_loop, daemon=True)
t.start()



import subprocess
import os
from ..execsteps import add_step, finish_step, ExecutionStatus, getscr
import re as regex

class Flags:
    Sudo = 0b001
    Strict = 0b010
    IgnoreErrors = 0b100
    DontPrintErrors = 0b1000
    HideStdout = 0b10000

def execute(windows_sh, linux_sh, flags: int = 0) -> subprocess.Popen:
    Sudo = (flags & 0b001) != 0
    Strict = (flags & 0b010) != 0
    IgnoreErrors = (flags & 0b100) != 0
    DontPrintErrors = (flags & 0b1000) != 0
    HideStdout = (flags & 0b10000) != 0

    if os.name == "nt":  # Windows
        shell = "powershell"
        args = ["-Command", windows_sh]
    else:  # Linux / MacOS
        shell = "bash"
        args = ["-c", linux_sh]
        if Sudo:
            code = f"sudo {code}"

    sid = add_step(f"Executing command (flags: s={Sudo},st={Strict},ie={IgnoreErrors},de={DontPrintErrors})")

    r = subprocess.Popen([shell, *args], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)

    stdout_data, stderr_data = r.communicate()

    if stdout_data and not HideStdout:
        add_step(stdout_data, ExecutionStatus.Stdout)
    if stderr_data and not DontPrintErrors:
        add_step(stderr_data, ExecutionStatus.Stderr)

    if r.returncode == 0 or IgnoreErrors:
        finish_step(sid, ExecutionStatus.CompletedSuccessfully)
    else:
        finish_step(sid, ExecutionStatus.Failed)
        if Strict:
            raise Exception(f"Command failed with return code {r.returncode}")

    return r



import subprocess
import os
from ..execsteps import add_step, finish_step, ExecutionStatus, getscr
import re as regex

class Flags:
    Recursive = 0b0001
    FolderContent = 0b0010
    FileContent = 0b0100
    RegEx = 0b1000

def _search(path: str, _filter: str, flags: int):
    Recursive = (flags & Flags.Recursive) != 0
    Folder = (flags & Flags.FolderContent) != 0
    File = (flags & Flags.FileContent) != 0
    RegEx = (flags & Flags.RegEx) != 0

    if (RegEx):
        f = regex.compile(_filter)
    else:
        f = (lambda x: (_filter in x))

    sid = add_step(f"Searching {path} (flags: r={Recursive},fo={Folder},fi={File},re={RegEx})")

    file_paths = []

    if not Recursive:
        if os.path.isfile(path):
            file_paths.append(path)
        else:
            for entry in os.listdir(path):
                full_path = os.path.join(path, entry)
                if os.path.isfile(full_path) and File:
                    file_paths.append(full_path)
                elif os.path.isdir(full_path) and Folder:
                    file_paths.append(full_path)
    else:
        def _walk(path):
            for root, dirs, files in os.walk(path):
                for file in files:
                    if not File:
                        continue
                    file_paths.append(os.path.join(root, file))
                for dir in dirs:
                    _walk(os.path.join(root, dir))
                    if Folder:
                        file_paths.append(os.path.join(root, dir))
        _walk(path)
                    
    matches = []
    
    for path in file_paths:
        if os.path.isfile(path):
            if Folder:
                if f(os.path.basename(path)):
                    matches.append({'type': 'file', 'matchtype': 'name', 'path': path})
            if File:
                try:
                    with open(path, 'r', errors='ignore') as file:
                        for i, line in enumerate(file):
                            if f(line):
                                matches.append({'type': 'file', 'matchtype': 'content', 'path': path, 'line': i + 1, 'text': line.strip()})
                except Exception as e:
                    add_step(f"Error reading file {path}: {e}", ExecutionStatus.Stderr)
        elif os.path.isdir(path) and Folder:
            if f(os.path.basename(path)):
                matches.append({'type': 'dir', 'matchtype': 'name', 'path': path})
        
    for match in matches:
        if match['type'] == 'file' and match['matchtype'] == 'content':
            add_step(f"Match: {match['path']} [{match['type']}] (Line {match['line']}): {match['text']}", ExecutionStatus.Stdout)
        elif match['type'] == 'file' and match['matchtype'] == 'name':
            add_step(f"Match: {match['path']} [{match['type']}]", ExecutionStatus.Stdout)
        elif match['type'] == 'dir':
            add_step(f"Match: {match['path']} [{match['type']}]", ExecutionStatus.Stdout)
    
    finish_step(sid, ExecutionStatus.CompletedSuccessfully)

    return matches


def search(path: str, _filter: str, flags: int) -> list | None:
    try:
        return _search(path, _filter, flags)
    except Exception as e:
        add_step(f"Error during search: {e}", ExecutionStatus.Error)
        return



import subprocess
import os
from ..execsteps import add_step, finish_step, ExecutionStatus, getscr

class Flags:
    Recursive = 0b01
    Force = 0b10
    IgnoreErrors = 0b1000

def rm(path: str, flags: int):
    Recursive = (flags & 0b01) != 0
    Force = (flags & 0b10) != 0
    ForceUseLinux = (flags & 0b100) != 0  # Debug Only
    IgnoreErrors = (flags & 0b1000) != 0
    if ForceUseLinux:
        print('ForceUseLinux active')

    sid = add_step(f"Deleting {path} (flags: r={Recursive},f={Force},fl={ForceUseLinux})")
    
    if os.name == "nt" and not ForceUseLinux:  # Windows
        args = []
        if Recursive: args.append("-Recurse")
        if Force: args.append("-Force")
        r = subprocess.Popen(["powershell", "-Command", f"Remove-Item {path} {' '.join(args)}"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
    else:
        args = []
        if Recursive: args.append("-r")
        if Force: args.append("-f")
        r =  subprocess.Popen(["rm", *args, path], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)

    stdout_data, stderr_data = r.communicate()

    if stdout_data:
        add_step(stdout_data, ExecutionStatus.Stdout)
    if stderr_data and not IgnoreErrors:
        add_step(stderr_data, ExecutionStatus.Stderr)

    if r.returncode == 0:
        finish_step(sid, ExecutionStatus.CompletedSuccessfully)
    else:
        if not IgnoreErrors:
            finish_step(sid, ExecutionStatus.Failed)
        else:
            finish_step(sid, ExecutionStatus.Warning)

    return r
